"""Test TotalConnectUser."""

import unittest

from total_connect_client.user import TotalConnectUser

USER_GOOD = {
    "UserID": "1234567",
    "Username": "username",
    "UserFeatureList": "Master=0,User Administration=0,Configuration Administration=0",
}

USER_BAD_MASTER = {
    "UserID": "1234567",
    "Username": "username",
    "UserFeatureList": "Master=1,User Administration=0,Configuration Administration=0",
}

USER_BAD_CONFIG = {
    "UserID": "1234567",
    "Username": "username",
    "UserFeatureList": "Master=0,User Administration=1,Configuration Administration=1",
}


class TestUser(unittest.TestCase):
    """Test TotalConnectUser."""

    def setUp(self):
        """Test setup."""
        self.user_good = TotalConnectUser(USER_GOOD)
        self.user_bad_master = TotalConnectUser(USER_BAD_MASTER)
        self.user_bad_config = TotalConnectUser(USER_BAD_CONFIG)

    def tearDown(self):
        """Tear down."""
        self.user_good = None
        self.user_bad_master = None
        self.user_bad_config = None

    def tests_security(self):
        """Test security_check()."""
        self.assertFalse(self.user_good.security_problem())
        self.assertTrue(self.user_bad_master.security_problem())
        self.assertTrue(self.user_bad_config.security_problem())



"""Exceptions raised by TotalConnectClient."""


class TotalConnectError(Exception):
    pass


class AuthenticationError(TotalConnectError):
    pass


class InvalidSessionError(AuthenticationError):
    pass


class BadResultCodeError(TotalConnectError):
    pass


class FeatureNotSupportedError(TotalConnectError):
    """The user hardware and/or account permissions do not support
    the requested feature.
    """


class RetryableTotalConnectError(TotalConnectError):
    """These errors are likely to resolve themselves if the action is retried.
    If an error requires some other action (such as reauthenticating an
    expired session) before being retried, it is not "retryable."
    """


class PartialResponseError(RetryableTotalConnectError):
    """Raised if the response is missing a section that it is always supposed to have.
    Because the TotalConnect servers are flaky, these are rather frequent.
    """


class UsercodeInvalid(TotalConnectError):
    """The provided usercode is invalid."""


class UsercodeUnavailable(TotalConnectError):
    """The TotalConnect usercode unavailable or invalid."""


class ServiceUnavailable(TotalConnectError):
    """The TotalConnect service is unavailable or unreachable."""


class FailedToBypassZone(TotalConnectError):
    """Failed to bypass zone because it is non-existent or it cannot be bypassed."""



"""Total Connect Partition."""

import logging
from typing import TYPE_CHECKING, Any, Final

from .const import PROJECT_URL, ArmingState, ArmType
from .exceptions import PartialResponseError, TotalConnectError

if TYPE_CHECKING:
    from .location import TotalConnectLocation

LOGGER: Final = logging.getLogger(__name__)


class TotalConnectPartition:
    """Partition class for Total Connect."""

    def __init__(self, details: dict[str, Any], parent: "TotalConnectLocation"):
        """Initialize Partition based on PartitionDetails."""
        self.parent: TotalConnectLocation = parent
        partition_id = details.get("PartitionID")
        if partition_id is None:
            raise TotalConnectError("PartitionID is required")
        self.partitionid: int = partition_id
        self.name: str | None = details.get("PartitionName")
        self.is_stay_armed: bool | None = details.get("IsStayArmed")
        self.is_fire_enabled: bool | None = details.get("IsFireEnabled")
        self.is_common_enabled: bool | None = details.get("IsCommonEnabled")
        self.is_locked: bool | None = details.get("IsLocked")
        self.is_new_partition: bool | None = details.get("IsNewPartition")
        self.is_night_stay_enabled: bool | None = details.get("IsNightStayEnabled")
        self.exit_delay_timer: int | None = details.get("ExitDelayTimer")
        self.arming_state: ArmingState  # Set by _update()
        self._update(details)

    def __str__(self) -> str:  # pragma: no cover
        """Return a string that is printable."""
        data = (
            f"PARTITION {self.partitionid} - {self.name}\n"
            f"  {self.arming_state}\n"
            f"  Stay armed: {self.is_stay_armed}\tFire enabled: {self.is_fire_enabled}\n"
            f"  Common enabled: {self.is_common_enabled}\tLocked: {self.is_locked}\n"
            f"  New: {self.is_new_partition}\tNight Stay enabled: {self.is_night_stay_enabled}\n"
            f"  Exit delay: {self.exit_delay_timer}\n"
        )

        return data

    def arm(self, arm_type: ArmType, usercode: str = "") -> None:
        """Arm the partition."""
        self.parent.arm(arm_type, self.partitionid, usercode)

    def disarm(self, usercode: str = "") -> None:
        """Disarm the partition."""
        self.parent.disarm(self.partitionid, usercode)

    def _update(self, info: dict[str, Any]) -> None:
        """Update partition state from PartitionInfo data."""
        astate = (info or {}).get("ArmingState")
        if astate is None:
            raise PartialResponseError("no ArmingState")
        try:
            self.arming_state = ArmingState(astate)
        except ValueError:
            LOGGER.error(
                f"unknown partition ArmingState {astate} in {info}: report at {PROJECT_URL}/issues"
            )
            raise TotalConnectError(f"unknown partition ArmingState {astate} in {info}") from None



"""Command-line tool to trigger Total Connect alarm systems.

WARNING: This script will trigger actual alarms on all locations!
Only use this for testing purposes in a controlled environment.

Usage:
    python -m total_connect_client.live.trigger username

Args:
    username: Total Connect username
"""

import getpass
import logging
import sys
from typing import Final, NoReturn

from ..client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)


def usage() -> NoReturn:
    """Print usage instructions."""
    print("usage:  username\n")
    print("WARNING: This will trigger actual alarms!")
    sys.exit()


def confirm_trigger() -> None:
    """Confirm that user really wants to trigger alarms."""
    print("WARNING: This will trigger actual alarms on ALL your locations!")
    print("This should only be used for testing purposes.")
    response = input("Are you sure you want to continue? (yes/NO): ")
    if response.lower() != "yes":
        print("Operation cancelled.")
        sys.exit(0)


if len(sys.argv) != 2:
    usage()

USERNAME: Final[str] = sys.argv[1]
PASSWORD: Final[str] = getpass.getpass()

TC: Final[TotalConnectClient] = TotalConnectClient(USERNAME, PASSWORD)

print("Authenticated with Total Connect")
confirm_trigger()

for location_id in TC.locations:
    location = TC.locations[location_id]
    print(f"Attempting to trigger location {location_id}")
    try:
        location.trigger()
        print(f"Successfully triggered alarm for location {location_id}")
    except Exception as e:
        print(f"Failed to trigger location {location_id}: {e}")
        logging.exception("Failed to trigger alarm for location %s", location_id)

print("Trigger operations completed. Check test.log for details.")
sys.exit()



"""Test TotalConnectClient."""

import requests_mock
from common import create_http_client
from const import RESPONSE_UNKNOWN, REST_RESULT_LOGOUT
from pytest import raises

from total_connect_client.const import HTTP_API_LOGOUT
from total_connect_client.exceptions import TotalConnectError


def tests_logout():
    """Test log_out."""
    client = create_http_client()
    assert client.is_logged_in() is True

    with requests_mock.Mocker() as rm:
        # first give an error
        rm.post(HTTP_API_LOGOUT, json=RESPONSE_UNKNOWN)
        with raises(TotalConnectError):
            client.log_out()
        assert client.is_logged_in() is True

        # then give success
        rm.post(HTTP_API_LOGOUT, json=REST_RESULT_LOGOUT)
        client.log_out()
        assert client.is_logged_in() is False



"""Command-line tool to arm Total Connect alarm systems.

Usage:
    python -m total_connect_client.live.arm username location1=usercode1,location2=usercode2 arm_type

Args:
    username: Total Connect username
    location_usercode_pairs: Comma-separated pairs of location=usercode
    arm_type: Numeric arm type (see ArmType enum)
"""

import getpass
import logging
import sys
from typing import Final, NoReturn

from ..client import TotalConnectClient
from ..const import ArmType

logging.basicConfig(filename="test.log", level=logging.DEBUG)


def usage() -> NoReturn:
    """Print usage instructions."""
    print("usage:  username location1=usercode1,location2=usercode2 arm_type \n")
    print("arm types:")
    for arm_type_enum in list(ArmType):
        print(f"  {arm_type_enum.name}: {arm_type_enum.value}")
    sys.exit()


if len(sys.argv) != 4:
    usage()

USERNAME: Final[str] = sys.argv[1]
USERCODES: Final[dict[str, str]] = dict(x.split("=") for x in sys.argv[2].split(","))
arm_type: Final[ArmType] = ArmType(int(sys.argv[3]))
PASSWORD: Final[str] = getpass.getpass()

TC: Final[TotalConnectClient] = TotalConnectClient(USERNAME, PASSWORD, USERCODES)

for location_id in TC.locations:
    location = TC.locations[location_id]
    location.arm(arm_type)

print(f"Armed: {arm_type}")

sys.exit()



"""Test TotalConnectDevice."""

from const import REST_RESULT_SESSION_DETAILS, SECURITY_DEVICE_ID

from total_connect_client.device import TotalConnectDevice

device_list = REST_RESULT_SESSION_DETAILS["SessionDetailsResult"]["Locations"][0]["DeviceList"]
# [0] is a ProA7 panel
# [1] is the built-in camera
# [2] is a Skybell doorbell


def tests_panel():
    """Test an alarm panel."""
    panel = TotalConnectDevice(device_list[0])
    assert panel.deviceid == SECURITY_DEVICE_ID
    assert panel.is_doorbell() is False



"""Total Connect Location."""

import logging
from typing import TYPE_CHECKING, Any, Final

from .const import PROJECT_URL, ArmingState, ArmType, _ResultCode, make_http_endpoint
from .device import TotalConnectDevice
from .exceptions import (
    FailedToBypassZone,
    FeatureNotSupportedError,
    PartialResponseError,
    TotalConnectError,
)
from .partition import TotalConnectPartition
from .zone import TotalConnectZone, ZoneStatus

if TYPE_CHECKING:
    from .client import TotalConnectClient

DEFAULT_USERCODE: Final[str] = "-1"

LOGGER: Final = logging.getLogger(__name__)


class TotalConnectLocation:
    """TotalConnectLocation class."""

    def __init__(self, location_info_basic: dict[str, Any], parent: "TotalConnectClient") -> None:
        """Initialize based on a 'LocationInfoBasic'."""
        self.location_id: int = location_info_basic["LocationID"]
        self.location_name: str = location_info_basic["LocationName"]
        self._photo_url: str = location_info_basic["PhotoURL"]
        self._module_flags: dict[str, str] = dict(
            x.split("=") for x in location_info_basic["LocationModuleFlags"].split(",")
        )
        self.security_device_id: str = location_info_basic["SecurityDeviceID"]
        self.parent: TotalConnectClient = parent
        self.ac_loss: bool | None = None
        self.low_battery: bool | None = None
        self.cover_tampered: bool | None = None
        self.last_updated_timestamp_ticks: int | None = None
        self.configuration_sequence_number: int | None = None
        self.arming_state: ArmingState = ArmingState.UNKNOWN
        self.partitions: dict[int, TotalConnectPartition] = {}
        self._partition_list: list[int] = []
        self.zones: dict[int, TotalConnectZone] = {}
        self.usercode: str = DEFAULT_USERCODE
        self.auto_bypass_low_battery: bool = False
        self._sync_job_id: str | None = None
        self._sync_job_state: int = 0

        dib = location_info_basic.get("DeviceList") or []
        tcdevs = [TotalConnectDevice(d) for d in dib]
        self.devices: dict[str, TotalConnectDevice] = {
            tcdev.deviceid: tcdev for tcdev in tcdevs if tcdev.deviceid is not None
        }

    def __str__(self) -> str:  # pragma: no cover
        """Return a text string that is printable."""
        data = (
            f"LOCATION {self.location_id} - {self.location_name}\n\n"
            f"PhotoURL: {self._photo_url}\n"
            f"SecurityDeviceID: {self.security_device_id}\n"
            f"AcLoss: {self.ac_loss}\n"
            f"LowBattery: {self.low_battery}\n"
            f"IsCoverTampered: {self.cover_tampered}\n"
            f"{self.arming_state}\n"
            f"LocationModuleFlags:\n"
        )

        for key, value in self._module_flags.items():
            data = data + f"  {key}: {value}\n"

        data = data + "\n"

        devices = f"DEVICES: {len(self.devices)}\n\n"
        for device in self.devices:
            devices = devices + str(self.devices[device]) + "\n"

        partitions = f"PARTITIONS: {len(self.partitions)}\n\n"
        for status in self.partitions.values():
            partitions += str(status) + "\n"

        zones = f"ZONES: {len(self.zones)}\n\n"
        for zone in self.zones.values():
            zones += str(zone)

        return data + devices + partitions + zones

    def get_panel_meta_data(self) -> None:
        """Get all meta data about the alarm panel."""
        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v3/locations/{self.location_id}/partitions/fullStatus"
            ),
            method="GET",
        )
        self.parent.raise_for_resultcode(result)

        self._update_status(result)
        self._update_partitions(result["PanelStatus"]["Partitions"])
        self._update_zones(result["PanelStatus"]["Zones"])

    def get_zone_details(self) -> None:
        """Get Zone details."""
        # 0 is the ListIdentifierID, whatever that might be
        result = self.parent.http_request(
            endpoint=make_http_endpoint(f"api/v1/locations/{self.location_id}/partitions/zones/0"),
            method="GET",
        )

        try:
            self.parent.raise_for_resultcode(result)
            self._update_zone_details(result)
        except FeatureNotSupportedError:
            LOGGER.warning(
                "getting Zone Details is a feature not supported by "
                "your Total Connect account or hardware"
            )

    def get_partition_details(self) -> None:
        """Get partition details for this location."""
        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v1/locations/{self.location_id}/devices/{self.security_device_id}/partitions/config"
            ),
            method="GET",
        )
        try:
            self.parent.raise_for_resultcode(result)
        except TotalConnectError:
            LOGGER.error(
                f"Could not get partition details for "
                f"device {self.security_device_id} at "
                f"location {self.location_id}."
                f"ResultCode: {result['ResultCode']}. "
                f"ResultData: {result['ResultData']}"
            )
            raise

        partition_details = (result or {}).get("Partitions")
        if not partition_details:
            raise PartialResponseError("no PartitionDetails", result)

        new_partition_list = []
        for partition in partition_details:
            new_partition = TotalConnectPartition(partition, self)
            self.partitions[new_partition.partitionid] = new_partition
            new_partition_list.append(new_partition.partitionid)

        self._partition_list = new_partition_list

    def is_low_battery(self) -> bool:
        """Return true if low battery."""
        return self.low_battery is True

    def is_ac_loss(self) -> bool:
        """Return true if AC loss."""
        return self.ac_loss is True

    def is_cover_tampered(self) -> bool:
        """Return true if cover is tampered."""
        return self.cover_tampered is True

    def set_usercode(self, usercode: str) -> bool:
        """Set the usercode. Return true if successful.

        NOTE: this only sets the code in this client for this session.
        It does not save the usercode into the panel.
        """
        try:
            self.validate_usercode(usercode)
        except TotalConnectError:
            return False

        self.usercode = usercode
        return True

    def validate_usercode(self, usercode: str) -> bool:
        """Return True if the usercode is in use at location."""
        response = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v1/account/users/current/validateUser/locations/{self.location_id}"
            ),
            method="POST",
            data={"userCode": int(usercode)},
        )

        try:
            self.parent.raise_for_resultcode(response)
        except TotalConnectError:
            LOGGER.error(
                f"Could not validate usercode ({usercode}) "
                f"for location {self.location_id}."
                f"Response: {response}. "
            )
            raise
        return bool(response["IsDuplicate"])

    def _build_partition_list(self, partition_id: int = 0) -> list[int]:
        """Build a list of partitions to use for arming/disarming."""
        if not partition_id:
            return self._partition_list

        if partition_id not in self.partitions:
            raise TotalConnectError(
                f"Partition {partition_id} does not exist at location {self.location_id}"
            )
        return [partition_id]

    def arm(self, arm_type: ArmType, partition_id: int = 0, usercode: str = "") -> None:
        """Arm the given partition. If not provided, arm the location.

        If no partition is given, arm all partitions.
        If no usercode given, use stored value."""
        assert isinstance(arm_type, ArmType)
        partition_list = self._build_partition_list(partition_id)
        usercode = usercode or self.usercode
        # treats usercode as int here, but str elsewhere
        usercode_int = int(usercode)

        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v3/locations/{self.location_id}/devices/{self.security_device_id}/partitions/arm"
            ),
            method="PUT",
            data={
                "armType": arm_type.value,
                "userCode": usercode_int,
                "partitions": partition_list,
            },
        )
        if _ResultCode.from_response(result) == _ResultCode.COMMAND_FAILED:
            LOGGER.warning("could not arm system; is a zone faulted?; is it already armed?")
        self.parent.raise_for_resultcode(result)
        LOGGER.info(f"ARMED({arm_type}) partitions {partition_list} at {self.location_id}")

    def disarm(self, partition_id: int = 0, usercode: str = "") -> None:
        """Disarm the system. If no partition given, disarm all of them."""
        if partition_id:
            # only check the partition
            if partition_id not in self.partitions:
                raise TotalConnectError(f"Requesting to disarm unknown partition {partition_id}")
            if (
                self.partitions[partition_id].arming_state.is_disarmed()
                or self.partitions[partition_id].arming_state.is_disarming()
            ):
                LOGGER.info(
                    f"Partition {partition_id} is already disarmed or in the process of disarming"
                )
                return
        else:
            # check the location
            if self.arming_state.is_disarmed() or self.arming_state.is_disarming():
                LOGGER.info(
                    f"Location {self.location_id} is already disarmed or in the process of disarming"
                )
                return

        partition_list = self._build_partition_list(partition_id)
        usercode = usercode or self.usercode
        # treats usercode as int here, but str elsewere
        usercode_int = int(usercode)

        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v3/locations/{self.location_id}/devices/{self.security_device_id}/partitions/disArm"
            ),
            method="PUT",
            data={"userCode": usercode_int, "partitions": partition_list},
        )
        self.parent.raise_for_resultcode(result)
        LOGGER.info(f"DISARMED partitions {partition_list} at location {self.location_id}")

    def zone_bypass(self, zone_id: int) -> None:
        """Bypass a zone."""
        self._bypass_zones([zone_id])

    def zone_bypass_all(self) -> None:
        """Bypass all faulted zones."""
        bypassable_faulted_zones = []
        for zone_id, zone in self.zones.items():
            if zone.is_faulted():
                if not zone.can_be_bypassed:
                    LOGGER.warning(
                        f"Zone {zone_id} ({zone.description}) is faulted but cannot be bypassed"
                    )
                    continue
                bypassable_faulted_zones.append(zone_id)

        self._bypass_zones(bypassable_faulted_zones)

    def _bypass_zones(self, zone_list: list[int]) -> None:
        """Bypass the given list of zones."""
        if not zone_list:
            LOGGER.info("Bypass request stopped because no zones are available to bypass")
            return

        # Validate zones before attempting bypass
        valid_zones = []
        for zone_id in zone_list:
            zone = self.zones.get(zone_id)
            if not zone:
                LOGGER.warning(f"Zone {zone_id} not found, skipping bypass")
                continue
            if not zone.can_be_bypassed:
                LOGGER.warning(f"Zone {zone_id} ({zone.description}) cannot be bypassed")
                continue
            if not zone.is_faulted():
                LOGGER.warning(f"Zone {zone_id} ({zone.description}) is not faulted, cannot bypass")
                continue
            valid_zones.append(zone_id)

        if not valid_zones:
            LOGGER.info("No valid zones found for bypass")
            return

        LOGGER.info(f"Attempting to bypass zones: {valid_zones}")

        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v1/locations/{self.location_id}/devices/{self.security_device_id}/bypass"
            ),
            method="PUT",
            data={"ZoneIds": valid_zones, "UserCode": int(self.usercode)},
        )

        self.parent.raise_for_resultcode(result)
        if _ResultCode.from_response(result) == _ResultCode.FAILED_TO_BYPASS_ZONE:
            raise FailedToBypassZone(f"Failed to bypass zone: {result}")

        # as of 12/31/2025
        # The API gives ResultCode 0 (success) but returns ZoneStatus 0 (normal) instead of 1 (bypassed)
        # but the next fullStatus panel update shows the zone as bypassed
        # so we can't rely on this response to update zone status

    def clear_bypass(self) -> None:
        """Clear all bypassed zones."""
        bypassed_zones = []
        for zone_id, zone in self.zones.items():
            if zone.is_bypassed():
                bypassed_zones.append(zone_id)

        if not bypassed_zones:
            LOGGER.info("Clear bypass request stopped because no zones are bypassed")
            return

        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v2/locations/{self.location_id}/devices/{self.security_device_id}/clearBypass"
            ),
            method="PUT",
            data={"userCode": int(self.usercode)},
        )

        self.parent.raise_for_resultcode(result)

    def zone_status(self, zone_id: int) -> ZoneStatus:
        """Get status of a zone."""
        zone = self.zones.get(zone_id)
        if not zone:
            raise TotalConnectError(f"zone {zone_id} does not exist")
        return zone.status

    def arm_custom(self, arm_type: ArmType) -> dict[str, Any]:
        """NOT OPERATIONAL YET."""
        raise TotalConnectError("arm_custom is not operational yet")

    def get_custom_arm_settings(self) -> dict[str, Any]:
        """NOT OPERATIONAL YET."""
        raise TotalConnectError("get_custom_arm_settings is not operational yet")

    def _update_zone_details(self, result: dict[str, Any]) -> None:
        """
        Update from ZoneStatusListEx_V1.

        ZoneStatusInfoWithPartitionId provides additional info for setting up zones.
        If we used TotalConnectZone._update() it would overwrite missing data with None.
        """
        zone_info = result["ZoneStatus"]["Zones"]
        if not zone_info:
            LOGGER.warning(
                "No zones found when starting TotalConnect. Try to sync your panel using the TotalConnect app or website."
            )
            LOGGER.debug(f"_update_zone_details result: {result}")
        else:
            for zonedata in zone_info:
                self.zones[zonedata["ZoneID"]] = TotalConnectZone(zonedata, self)

    def _update_status(self, result: dict[str, Any]) -> None:
        """Update from result."""
        data = (result or {}).get("PanelStatus")
        if not data:
            raise PartialResponseError("no PanelStatus", result)

        self.ac_loss = data.get("IsInACLoss")
        self.low_battery = data.get("IsInLowBattery")
        self.cover_tampered = data.get("IsCoverTampered")
        self.last_updated_timestamp_ticks = data.get("LastUpdatedTimestampTicks")
        self.configuration_sequence_number = data.get("ConfigurationSequenceNumber")

        # TODO: new resposne structure has a lot more data than we use here
        # and some fields are provided twice...are we using the right tones?

        astate = result.get("ArmingState")
        if not astate:
            raise PartialResponseError("no ArmingState", result)
        try:
            self.arming_state = ArmingState(astate)
        except ValueError:
            LOGGER.error(
                f"unknown location ArmingState {astate} in {result}: please report at {PROJECT_URL}/issues"
            )
            raise TotalConnectError(f"unknown location ArmingState {astate} in {result}") from None

    def _update_partitions(self, partitions: list[dict[str, Any]]) -> None:
        """Update partition info from Partitions."""
        # loop through partitions and update
        # NOTE: do not use keys because they don't line up with PartitionID
        for partition in partitions:
            if "PartitionID" not in partition:
                raise PartialResponseError("no PartitionID", partitions)
            partition_id = int(partition["PartitionID"])
            if partition_id in self.partitions:
                self.partitions[partition_id]._update(partition)
            else:
                LOGGER.warning(f"Update provided for unknown partion {partition_id}")

    def _update_zones(self, zones: list[dict[str, Any]]) -> None:
        """Update zone info from Zones."""
        if not zones:
            LOGGER.error("no zones found: sync your panel using TotalConnect app or website")
            raise TotalConnectError("no zones found: panel sync required")

        for zonedata in zones:
            zone_id = int(zonedata["ZoneID"])
            zone = self.zones.get(zone_id)
            if zone:
                zone._update(zonedata)
            else:
                zone = TotalConnectZone(zonedata, self)
                self.zones[zone_id] = zone

            if zone.is_low_battery() and zone.can_be_bypassed and self.auto_bypass_low_battery:
                self.zone_bypass(zone_id)

    def sync_panel(self) -> None:
        """Syncronize the panel with the TotalConnect server."""
        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v1/locations/{self.location_id}/devices/security/synchronize"
            ),
            method="POST",
            data={"userCode": self.usercode},
        )
        self.parent.raise_for_resultcode(result)
        self._sync_job_id = result.get("JobID")
        # Successful request so assume state is in progress
        LOGGER.info(f"Started sync of panel for location {self.location_id}")

    def get_cameras(self) -> None:
        """Get cameras for the location."""
        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v1/locations/{self.location_id}/GetLocationAllCameraListEx"
            ),
            method="GET",
        )
        self.parent.raise_for_resultcode(result)

        if "AccountAllCameraList" not in result:
            LOGGER.info(f"No cameras found for location {self.location_id}")
            return

        camera_list = result["AccountAllCameraList"]

        if "WiFiDoorbellList" in camera_list:
            self._get_doorbell(camera_list["WiFiDoorbellList"])

        if "UnicornList" in camera_list:
            self._get_unicorn(camera_list["UnicornList"])

        if "VideoPirList" in camera_list:
            self._get_video(camera_list["VideoPirList"])

        # TODO: look at Gen5DoorbellList and DoorBellList and lots of other info available

    def _get_doorbell(self, data: dict[str, Any]) -> None:
        """Find doorbell info."""
        if not data or "WiFiDoorbellsList" not in data:
            return

        doorbells = data["WiFiDoorbellsList"]
        if "WiFiDoorBellInfo" not in doorbells:
            return

        doorbell_list = doorbells["WiFiDoorBellInfo"]
        for doorbell in doorbell_list:
            id = doorbell["DeviceID"]
            if id in self.devices:
                self.devices[id].doorbell_info = doorbell

    def _get_unicorn(self, data: dict[str, Any]) -> None:
        """Find uniforn info."""
        if not data or "UnicornList" not in data:
            return

        unicorns = data["UnicornsList"]
        if "UnicornInfo" not in unicorns:
            return

        unicorn_list = unicorns["UnicornInfo"]
        for unicorn in unicorn_list:
            id = unicorn["DeviceID"]
            if id in self.devices:
                self.devices[id].unicorn_info = unicorn

    def _get_video(self, data: dict[str, Any]) -> None:
        """Get video for the location."""
        if not data or "VideoPirInfo" not in data:
            return

        info = data["VideoPirInfo"]

        for video in info:
            id = video["DeviceID"]
            if id in self.devices:
                self.devices[id].video_info = video

    def trigger(self) -> None:
        """Trigger the alarm (experimental)."""
        result = self.parent.http_request(
            endpoint=make_http_endpoint(
                f"api/v1/locations/{self.location_id}/devices/{self.security_device_id}/RemotePanicAlarm"
            ),
            method="POST",
        )
        LOGGER.debug(f"trigger result:\n{result}")
        self.parent.raise_for_resultcode(result)
        LOGGER.info(f"Triggered alarm at {self.location_id}")



"""Command-line tool to test zone bypass functionality.

This script tests the complete bypass workflow:
1. Identifies faulted zones in the alarm system
2. Attempts to bypass all faulted zones
3. Waits 10 seconds for the system to process
4. Attempts to clear all bypassed zones

Usage:
    python -m total_connect_client.live.bypass username location1=usercode1,location2=usercode2

Args:
    username: Total Connect username
    location_usercode_pairs: Comma-separated pairs of location=usercode
"""

import getpass
import logging
import sys
import time
from typing import Final, NoReturn

from ..client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)

BYPASS_WAIT_TIME: Final[int] = 10


def usage() -> NoReturn:
    """Print usage instructions."""
    print(
        "usage:  python -m total_connect_client.live.bypass username location1=usercode1,location2=usercode2\n"
    )
    print("This script will:")
    print("1. Try to bypass all faulted zones")
    print("2. Wait for 10 seconds")
    print("3. Try to clear all bypassed zones")
    sys.exit()


if len(sys.argv) != 3:
    usage()

USERNAME: Final[str] = sys.argv[1]
USERCODES: Final[dict[str, str]] = dict(x.split("=") for x in sys.argv[2].split(","))
PASSWORD: Final[str] = getpass.getpass()

TC: Final[TotalConnectClient] = TotalConnectClient(USERNAME, PASSWORD, USERCODES)

for location_id in TC.locations:
    location = TC.locations[location_id]
    print(f"\n=== Testing bypass for location {location_id} ({location.location_name}) ===")

    # Get current zone status
    location.get_zone_details()

    print(f"Found {len(location.zones)} zones")

    # Show faulted zones
    faulted_zones: list[int] = []
    for zone_id, zone in location.zones.items():
        if zone.is_faulted():
            faulted_zones.append(zone_id)
            print(f"Zone {zone_id} ({zone.description}) is faulted")

    if not faulted_zones:
        print("No faulted zones found. Cannot test bypass functionality.")
        continue

    print(f"\nAttempting to bypass {len(faulted_zones)} faulted zones: {faulted_zones}")

    try:
        # Bypass all faulted zones
        location.zone_bypass_all()
        print("Successfully initiated bypass for all faulted zones")

        # Wait for system to process
        print(f"Waiting {BYPASS_WAIT_TIME} seconds...")
        time.sleep(BYPASS_WAIT_TIME)

        # Refresh zone status
        location.get_zone_details()

        # Show bypassed zones
        bypassed_zones: list[int] = []
        for zone_id, zone in location.zones.items():
            if zone.is_bypassed():
                bypassed_zones.append(zone_id)
                print(f"Zone {zone_id} ({zone.description}) is bypassed")

        print(f"\nAttempting to clear bypass for {len(bypassed_zones)} bypassed zones")

        # Clear all bypassed zones
        location.clear_bypass()
        print("Successfully cleared all bypassed zones")

    except Exception as e:
        print(f"Error during bypass test: {e}")
        logging.exception("Error during bypass test")

print("\nBypass test completed. Check test.log for detailed debug information.")



"""Tests TotalConnectLocation."""

from unittest.mock import Mock

import requests_mock
from common import create_http_client
from const import (
    LOCATION_ID,
    PANEL_STATUS_ARMED_AWAY,
    PANEL_STATUS_DISARMED,
    RESPONSE_DISARM_SUCCESS,
    RESPONSE_UNKNOWN,
    REST_RESULT_FULL_STATUS,
    REST_RESULT_PARTITIONS_CONFIG,
    REST_RESULT_PARTITIONS_ZONES,
    REST_RESULT_SESSION_DETAILS,
    REST_RESULT_VALIDATE_USER_LOCATIONS,
)
from pytest import raises

from total_connect_client.const import ArmingState, ArmType, make_http_endpoint
from total_connect_client.exceptions import (
    FeatureNotSupportedError,
    PartialResponseError,
    TotalConnectError,
)
from total_connect_client.location import TotalConnectLocation

RESULT_LOCATION = REST_RESULT_SESSION_DETAILS["SessionDetailsResult"]["Locations"][0]
result_num_zones = len(REST_RESULT_PARTITIONS_ZONES["ZoneStatus"]["Zones"])


def tests_location_basic():
    """Tests basic location info."""
    location = TotalConnectLocation(RESULT_LOCATION, Mock())
    assert location.location_id == LOCATION_ID
    assert location.is_ac_loss() is False
    assert location.is_cover_tampered() is False
    assert location.is_low_battery() is False


def tests_get_partition_details():
    """Test get_partition_details function."""

    client = Mock()
    location = TotalConnectLocation(RESULT_LOCATION, client)
    assert len(location.partitions) == 0

    # first an error
    client.http_request.return_value = RESPONSE_UNKNOWN
    client.raise_for_resultcode.side_effect = TotalConnectError()
    with raises(TotalConnectError):
        location.get_partition_details()
    assert len(location.partitions) == 0

    # now with partial data
    client.http_request.return_value = {}
    client.raise_for_resultcode.side_effect = None
    with raises(PartialResponseError):
        location.get_partition_details()
    assert len(location.partitions) == 0

    # now it works
    client.http_request.return_value = REST_RESULT_PARTITIONS_CONFIG
    location.get_partition_details()
    assert len(location.partitions) == 1


def tests_get_zone_details():
    """Test get_zone_details function."""

    client = Mock()
    client.http_request.return_value = REST_RESULT_PARTITIONS_ZONES
    client.raise_for_resultcode.return_value = None

    location = TotalConnectLocation(RESULT_LOCATION, client)
    assert len(location.zones) == 0

    # first an error
    client.raise_for_resultcode.side_effect = FeatureNotSupportedError()
    location.get_zone_details()
    assert len(location.zones) == 0

    # now it should work
    client.raise_for_resultcode.side_effect = None
    location.get_zone_details()
    assert len(location.zones) == result_num_zones


def tests_get_panel_metadata():
    """Test status updates."""

    client = Mock()
    client.http_request.return_value = REST_RESULT_PARTITIONS_ZONES
    client.raise_for_resultcode.return_value = None

    location = TotalConnectLocation(RESULT_LOCATION, client)
    location.get_zone_details()
    assert len(location.zones) == result_num_zones
    assert location.arming_state == ArmingState.UNKNOWN

    client.http_request.return_value = REST_RESULT_FULL_STATUS
    location.get_panel_meta_data()
    assert location.arming_state == ArmingState.DISARMED_ZONE_FAULTED


def tests_usercode():
    """Test usercode fuctions."""
    client = Mock()
    location = TotalConnectLocation(RESULT_LOCATION, client)

    # first an error
    client.http_request.return_value = RESPONSE_UNKNOWN
    client.raise_for_resultcode.side_effect = TotalConnectError()
    assert location.set_usercode("1234") is False

    client.raise_for_resultcode.side_effect = None
    client.http_request.return_value = REST_RESULT_VALIDATE_USER_LOCATIONS
    assert location.set_usercode("1234") is True


def tests_disarm():
    """Test disarm."""
    client = create_http_client(PANEL_STATUS_ARMED_AWAY)
    location = client.locations[LOCATION_ID]
    assert location.arming_state.is_armed()

    with requests_mock.Mocker() as rm:
        rm.put(
            make_http_endpoint(
                f"api/v3/locations/{location.location_id}/devices/{location.security_device_id}/partitions/disArm"
            ),
            json=RESPONSE_DISARM_SUCCESS,
        )

        # try to disarm a non-existent partition
        with raises(TotalConnectError):
            location.disarm(999, "1234")

        # now should work
        location.disarm(1, "1234")

        rm.get(
            make_http_endpoint(f"api/v3/locations/{location.location_id}/partitions/fullStatus"),
            json=PANEL_STATUS_DISARMED,
        )
        location.get_panel_meta_data()
        assert location.arming_state.is_disarmed()

        # now should do nothing because already disarmed
        location.disarm(1, "1234")
        assert location.arming_state.is_disarmed()

        # now try just the location
        rm.get(
            make_http_endpoint(f"api/v3/locations/{location.location_id}/partitions/fullStatus"),
            json=PANEL_STATUS_ARMED_AWAY,
        )
        location.get_panel_meta_data()
        assert location.arming_state.is_armed()

        rm.put(
            make_http_endpoint(
                f"api/v3/locations/{location.location_id}/devices/{location.security_device_id}/partitions/disArm"
            ),
            json=RESPONSE_DISARM_SUCCESS,
        )
        location.disarm(usercode="1234")

        rm.get(
            make_http_endpoint(f"api/v3/locations/{location.location_id}/partitions/fullStatus"),
            json=PANEL_STATUS_DISARMED,
        )
        location.get_panel_meta_data()
        assert location.arming_state.is_disarmed()

        # now should do nothing because already disarmed
        location.disarm(usercode="1234")
        assert location.arming_state.is_disarmed()


def tests_arm():
    """Test arm."""
    client = create_http_client(PANEL_STATUS_DISARMED)
    location = client.locations[LOCATION_ID]
    assert location.arming_state.is_disarmed()

    with requests_mock.Mocker() as rm:
        rm.put(
            make_http_endpoint(
                f"api/v3/locations/{location.location_id}/devices/{location.security_device_id}/partitions/arm"
            ),
            json=RESPONSE_DISARM_SUCCESS,
        )

        # try to arm a non-existent partition
        with raises(TotalConnectError):
            location.arm(partition_id=999, usercode="1234", arm_type=ArmType.AWAY)

        # now should work
        location.arm(partition_id=1, usercode="1234", arm_type=ArmType.AWAY)

        # now just the location should work
        location.arm(usercode="1234", arm_type=ArmType.AWAY)



"""Total Connect Zone."""

import logging
from enum import Enum, IntFlag
from typing import TYPE_CHECKING, Any, Final

from .const import PROJECT_URL
from .exceptions import TotalConnectError

if TYPE_CHECKING:
    from .location import TotalConnectLocation

LOGGER: Final = logging.getLogger(__name__)


class ZoneStatus(IntFlag):
    """Class to represent ZoneStatus."""

    NORMAL = 0
    BYPASSED = 1
    FAULT = 2  # only returned when Sensor Events are enabled in account
    TROUBLE = 8  # is also Tampered
    TAMPER = 16  # Tamper for ProA7, see #176
    COMMUNICATION_FAILURE = 32  # see #191
    LOW_BATTERY = 64
    TRIGGERED = 256
    KNOWN = (
        NORMAL
        | BYPASSED
        | FAULT
        | TROUBLE
        | TAMPER
        | COMMUNICATION_FAILURE
        | LOW_BATTERY
        | TRIGGERED
    )


class ZoneType(Enum):
    """
    These are the "standard" Honeywell zone types. However some panels such
    as the Lynx 7000 report most security zones as zone type 3.

    https://www.alarmliquidators.com/content/Vista%2021IP-%20Programming%20Guide.pdf
    http://techresource.online/training/ssnw/honeywell/zone-types
    """

    SECURITY = 0  # for Vista, zone type 0 is not used
    ENTRY_EXIT1 = 1  # starts countdown timer #1
    ENTRY_EXIT2 = 2  # like ENTRY_EXIT1 but uses timer #2
    PERIMETER = 3  # zone type 3 usually triggers an immediate alarm...
    PROA7_SECURITY = 3  # but some panels like Lynx 7000 report timed zones as 3
    INTERIOR_FOLLOWER = 4  # inactive when armed STAY
    TROUBLE_ALARM = 5  # trouble by day, alarm by night
    SILENT_24HR = 6  # 24-hour silent alarm (often used for police/hold-up)
    AUDIBLE_24HR = 7  # 24-hour audible alarm (often used for police)
    AUX_24HR = 8  # no local siren but keypad beeps (often used for medical)
    FIRE_SMOKE = 9
    INTERIOR_DELAY = 10  # inactive when armed STAY, otherwise like ENTRY_EXIT1
    MONITOR = 12  # e.g. temperature or flood
    CARBON_MONOXIDE = 14
    PROA7_MEDICAL = 15
    FIRE_W_VERIFICATION = 16  # must trigger twice to cause an alarm
    RF_ARM_STAY = 20  # keyfob
    RF_ARM_AWAY = 21  # keyfob
    RF_DISARM = 22  # keyfob
    NO_ALARM_RESPONSE = 23  # per Vista docs
    SILENT_BURGLARY = 24  # per Vista docs
    LYRIC_KEYPAD = 50
    PROA7_GARAGE_MONITOR = 53  # issue #167
    KEYSWITCH = 77  # per Vista20P docs
    AAV_MONITOR = 81  # per Vista20P docs
    LYRIC_LOCAL_ALARM = 89

    # According to the VISTA docs, these can be programmed via downloader
    # software or from a keypad using data fields *182-*185

    VISTA_CONFIGURABLE_90 = 90
    VISTA_CONFIGURABLE_91 = 91
    VISTA_CONFIGURABLE_92 = 92
    VISTA_CONFIGURABLE_93 = 93


class TotalConnectZone:
    """Do not create instances of this class yourself."""

    def __init__(self, zone: dict[str, Any], parent_location: "TotalConnectLocation") -> None:
        """Initialize."""
        zone_id = zone.get("ZoneID")
        if zone_id is None:
            raise TotalConnectError("ZoneID is required")
        self.zoneid: int = zone_id
        self._parent_location: TotalConnectLocation = parent_location
        self.partition: int = 0
        self.status: ZoneStatus = ZoneStatus.NORMAL
        self.zone_type_id: ZoneType | int | None = None
        self.can_be_bypassed: bool | None = None
        self.battery_level: int | None = None
        self.signal_strength: int | None = None
        self.sensor_serial_number: str | None = None
        self.loop_number: int | None = None
        self.response_type: str | None = None
        self.alarm_report_state: str | None = None
        self.supervision_type: str | None = None
        self.chime_state: int | None = None
        self.device_type: int | None = None
        self._unknown_type_reported: bool = False
        self.description: str | None  # Set by _update()
        self._update(zone)

    def __str__(self) -> str:  # pragma: no cover
        """Return a string that is printable."""
        return (
            f"Zone {self.zoneid} - {self.description}\n"
            f"  Partition: {self.partition}\t\t"
            f"Zone Type: {self.zone_type_id}\t"
            f"CanBeBypassed: {self.can_be_bypassed}\t"
            f"Status: {self.status}\n"
            f"  Battery Level: {self.battery_level}\t"
            f"Signal Stength: {self.signal_strength}\n"
            f"  Serial Number: {self.sensor_serial_number}\t"
            f"Loop: {self.loop_number}\t"
            f"Response Type: {self.response_type}\n"
            f"  Supervision Type: {self.supervision_type}\t"
            f"Alarm Report State: {self.alarm_report_state}\n"
            f"  Chime State: {self.chime_state}\t"
            f"Device Type: {self.device_type}\n\n"
        )

    def is_bypassed(self) -> bool:
        """Return true if the zone is bypassed."""
        return self.status & ZoneStatus.BYPASSED > 0

    def is_faulted(self) -> bool:
        """Return true if the zone is faulted."""
        return self.status & ZoneStatus.FAULT > 0

    def is_tampered(self) -> bool:
        """Return true if zone is tampered."""
        return (self.status & ZoneStatus.TROUBLE > 0) or (self.status & ZoneStatus.TAMPER > 0)

    def is_low_battery(self) -> bool:
        """Return true if low battery."""
        return self.status & ZoneStatus.LOW_BATTERY > 0

    def is_troubled(self) -> bool:
        """Return true if zone is troubled."""
        return self.status & ZoneStatus.TROUBLE > 0

    def is_triggered(self) -> bool:
        """Return true if zone is triggered."""
        return self.status & ZoneStatus.TRIGGERED > 0

    def is_type_button(self) -> bool:
        """Return true if zone is a button."""
        # as seen so far, any security zone that cannot be bypassed
        # is a button on a panel
        if self.is_type_security() and not self.can_be_bypassed:
            return True

        if self.zone_type_id in (
            ZoneType.PROA7_MEDICAL,
            ZoneType.AUDIBLE_24HR,
            ZoneType.SILENT_24HR,
            ZoneType.RF_ARM_STAY,
            ZoneType.RF_ARM_AWAY,
            ZoneType.RF_DISARM,
        ):
            return True

        return False

    def is_type_security(self) -> bool:
        """Return true if zone type is security."""
        return self.zone_type_id in (
            ZoneType.SECURITY,
            ZoneType.ENTRY_EXIT1,
            ZoneType.ENTRY_EXIT2,
            ZoneType.PERIMETER,
            ZoneType.INTERIOR_FOLLOWER,
            ZoneType.TROUBLE_ALARM,
            ZoneType.SILENT_24HR,
            ZoneType.AUDIBLE_24HR,
            ZoneType.INTERIOR_DELAY,
            ZoneType.LYRIC_LOCAL_ALARM,
            ZoneType.PROA7_GARAGE_MONITOR,
        )

    def is_type_motion(self) -> bool:
        """Return true if zone type is motion."""
        return self.zone_type_id == ZoneType.INTERIOR_FOLLOWER

    def is_type_fire(self) -> bool:
        """Return true if zone type is fire or smoke."""
        return self.zone_type_id == ZoneType.FIRE_SMOKE

    def is_type_temperature(self) -> bool:
        """Return true if zone monitors the temperature."""
        return self.zone_type_id == ZoneType.MONITOR

    def is_type_carbon_monoxide(self) -> bool:
        """Return true if zone type is carbon monoxide."""
        return self.zone_type_id == ZoneType.CARBON_MONOXIDE

    def is_type_medical(self) -> bool:
        """Return true if zone type is medical."""
        return self.zone_type_id == ZoneType.PROA7_MEDICAL

    def is_type_keypad(self) -> bool:
        """Return true if zone type is keypad."""
        return self.zone_type_id == ZoneType.LYRIC_KEYPAD

    def _update(self, zone: dict[str, Any]) -> None:
        """Update zone state from zone data."""
        if not zone:
            raise TotalConnectError("Missing data in zone update")

        zid = zone.get("ZoneID")
        if not self.zoneid == zid:
            raise TotalConnectError("Zone ID mismatch")

        self.description = zone.get("ZoneDescription")
        # ZoneInfo gives 'PartitionID' but
        # ZoneStatusInfoWithPartitionId gives 'PartitionId'
        partition_value = zone.get("PartitionId") or zone.get("PartitionID")
        self.partition = partition_value if partition_value is not None else 0

        try:
            zone_status_value = zone.get("ZoneStatus")
            if zone_status_value is None:
                raise ValueError("ZoneStatus is required")
            status = ZoneStatus(zone_status_value)
            if status & ~ZoneStatus.KNOWN > 0:
                LOGGER.warning(
                    f"unknown ZoneStatus {zone.get('ZoneStatus')} in {zone}: please report at {PROJECT_URL}/issues"
                )
            self.status = status

        except ValueError:
            LOGGER.error(
                f"invalid ZoneStatus {zone.get('ZoneStatus')} in {zone}: please report at {PROJECT_URL}/issues"
            )
            raise TotalConnectError(f"unknown ZoneStatus in {zone}") from None

        self.can_be_bypassed = zone.get("CanBeBypassed")

        try:
            zid = zone.get("ZoneTypeId", self.zone_type_id)
            # TODO: if zid is None should we raise PartialResponseError?
            self.zone_type_id = None if zid is None else ZoneType(zid)
        except ValueError:
            # if we get an unknown ZoneType we do not raise an exception, because
            # we know there are more zone types than we have in our enum, and
            # having an unknown ZoneType doesn't keep us from doing our work
            if not self._unknown_type_reported:
                LOGGER.info(
                    f"unknown ZoneType {zid} in {zone}: please report at {PROJECT_URL}/issues"
                )
                self._unknown_type_reported = True
            self.zone_type_id = zid

        self.battery_level = zone.get("Batterylevel", self.battery_level)
        self.signal_strength = zone.get("Signalstrength", self.signal_strength)
        info = zone.get("zoneAdditionalInfo")
        if info:
            self.sensor_serial_number = info.get("SensorSerialNumber")
            self.loop_number = info.get("LoopNumber")
            self.response_type = info.get("ResponseType")
            self.alarm_report_state = info.get("AlarmReportState")
            self.supervision_type = info.get("ZoneSupervisionType")
            self.chime_state = info.get("ChimeState")
            self.device_type = info.get("DeviceType")

    def _mark_as_bypassed(self) -> None:
        """Set is_bypassed status."""
        self.status |= ZoneStatus.BYPASSED

    def bypass(self) -> None:
        """Bypass the zone."""
        if self.can_be_bypassed:
            self._parent_location.zone_bypass(self.zoneid)



"""Total Connect Client constants."""

import urllib.parse
from enum import Enum
from typing import Any, Final

from .exceptions import BadResultCodeError


class ArmType(Enum):
    """Represent ArmingType."""

    AWAY = 0
    STAY = 1
    STAY_INSTANT = 2
    AWAY_INSTANT = 3
    STAY_NIGHT = 4


class ArmingState(Enum):
    """Represent ArmingState."""

    DISARMED = 10200
    DISARMED_BYPASS = 10211
    DISARMED_ZONE_FAULTED = 10214  # seems to apply to location, not to partition.  See issue #144

    ARMED_AWAY = 10201
    ARMED_AWAY_BYPASS = 10202
    ARMED_STAY = 10203
    ARMED_STAY_OTHER = 10226
    ARMED_STAY_PROA7 = 10230  # issue #173
    ARMED_STAY_BYPASS_PROA7 = 10231  # issue #177
    ARMED_STAY_INSTANT_PROA7 = 10232  # issue #177
    ARMED_STAY_INSTANT_BYPASS_PROA7 = 10233  # issue #177
    ARMED_STAY_BYPASS = 10204
    ARMED_AWAY_INSTANT = 10205
    ARMED_AWAY_INSTANT_BYPASS = 10206
    ARMED_STAY_INSTANT = 10209
    ARMED_STAY_INSTANT_BYPASS = 10210
    ARMED_STAY_NIGHT = 10218
    ARMED_STAY_NIGHT_BYPASS_PROA7 = 10219  # issue #177
    ARMED_STAY_NIGHT_INSTANT_PROA7 = 10220  # issue #177
    ARMED_STAY_NIGHT_INSTANT_BYPASS_PROA7 = 10221  # issue #177
    ARMED_CUSTOM_BYPASS = 10223

    ALARMING = 10207
    ALARMING_FIRE_SMOKE = 10212
    ALARMING_CARBON_MONOXIDE = 10213
    ALARMING_CARBON_MONOXIDE_PROA7 = 10215

    ARMING = 10307
    DISARMING = 10308

    UNKNOWN = 0

    def is_arming(self) -> bool:
        """Return true if the system is in the process of arming."""
        return self == ArmingState.ARMING

    def is_disarming(self) -> bool:
        """Return true if the system is in the process of disarming."""
        return self == ArmingState.DISARMING

    def is_pending(self) -> bool:
        """Return true if the system is pending an action."""
        return self.is_disarming() or self.is_arming()

    def is_disarmed(self) -> bool:
        """Return True if the system is disarmed."""
        return self in (
            ArmingState.DISARMED,
            ArmingState.DISARMED_BYPASS,
            ArmingState.DISARMED_ZONE_FAULTED,
        )

    def is_armed_away(self) -> bool:
        """Return True if the system is armed away in any way."""
        return self in (
            ArmingState.ARMED_AWAY,
            ArmingState.ARMED_AWAY_BYPASS,
            ArmingState.ARMED_AWAY_INSTANT,
            ArmingState.ARMED_AWAY_INSTANT_BYPASS,
        )

    def is_armed_custom_bypass(self) -> bool:
        """Return True if the system is armed custom bypass in any way."""
        return self == ArmingState.ARMED_CUSTOM_BYPASS

    def is_armed_home(self) -> bool:
        """Return True if the system is armed home/stay in any way."""
        return self in (
            ArmingState.ARMED_STAY,
            ArmingState.ARMED_STAY_PROA7,
            ArmingState.ARMED_STAY_BYPASS,
            ArmingState.ARMED_STAY_BYPASS_PROA7,
            ArmingState.ARMED_STAY_INSTANT,
            ArmingState.ARMED_STAY_INSTANT_PROA7,
            ArmingState.ARMED_STAY_INSTANT_BYPASS,
            ArmingState.ARMED_STAY_INSTANT_BYPASS_PROA7,
            ArmingState.ARMED_STAY_NIGHT,
            ArmingState.ARMED_STAY_NIGHT_BYPASS_PROA7,
            ArmingState.ARMED_STAY_NIGHT_INSTANT_PROA7,
            ArmingState.ARMED_STAY_NIGHT_INSTANT_BYPASS_PROA7,
            ArmingState.ARMED_STAY_OTHER,
        )

    def is_armed_night(self) -> bool:
        """Return True if the system is armed night in any way."""
        return self in (
            ArmingState.ARMED_STAY_NIGHT,
            ArmingState.ARMED_STAY_NIGHT_BYPASS_PROA7,
            ArmingState.ARMED_STAY_NIGHT_INSTANT_PROA7,
            ArmingState.ARMED_STAY_NIGHT_INSTANT_BYPASS_PROA7,
            # per #240 STAY_INSTANT and STAY_INSTANT_BYPASS are actually Night
            ArmingState.ARMED_STAY_INSTANT,
            ArmingState.ARMED_STAY_INSTANT_BYPASS,
        )

    def is_armed(self) -> bool:
        """Return True if the system is armed in any way."""
        return (
            self.is_armed_away()
            or self.is_armed_home()
            or self.is_armed_night()  # noqa: W504
            or self.is_armed_custom_bypass()
        )

    def is_triggered_police(self) -> bool:
        """Return True if the system is triggered for police or medical."""
        return self == ArmingState.ALARMING

    def is_triggered_fire(self) -> bool:
        """Return True if the system is triggered for fire or smoke."""
        return self == ArmingState.ALARMING_FIRE_SMOKE

    def is_triggered_gas(self) -> bool:
        """Return True if the system is triggered for carbon monoxide."""
        return self in (
            ArmingState.ALARMING_CARBON_MONOXIDE,
            ArmingState.ALARMING_CARBON_MONOXIDE_PROA7,
        )

    def is_triggered(self) -> bool:
        """Return True if the system is triggered in any way."""
        return self.is_triggered_fire() or self.is_triggered_gas() or self.is_triggered_police()


class _ResultCode(Enum):
    """Represent ResultCode.

    As suggested by the leading underscore, this class is not used by
    callers of the API.
    """

    @staticmethod
    def from_response(response_dict: dict[str, Any]) -> "_ResultCode":
        """Create a _ResultCode from an API response dictionary.

        Args:
            response_dict: Dictionary containing API response with ResultCode or error field

        Returns:
            _ResultCode enum value corresponding to the response

        Raises:
            BadResultCodeError: If the result code is unknown
        """
        value = response_dict.get("ResultCode") or response_dict.get("error")
        if not value:
            return _ResultCode.SUCCESS
        try:
            return _ResultCode(int(value))
        except ValueError:
            raise BadResultCodeError(f"unknown result code {value}", response_dict) from None

    SUCCESS = 0
    ARM_SUCCESS = 4500
    DISARM_SUCCESS = 4500
    SESSION_INITIATED = 4500

    BAD_USER_OR_PASSWORD = -50004
    INVALID_SESSIONID = -30002
    FAILED_TO_BYPASS_ZONE = -4504
    COMMAND_FAILED = -4502
    USER_CODE_UNAVAILABLE = -4114
    CANNOT_CONNECT = -4108  # see #262
    USER_CODE_INVALID = -4106
    FAILED_TO_CONNECT = -4104

    # Invalid Parameter returned when requesting SyncPanelStatus
    # using non-existant JobID.
    INVALID_PARAMETER = -501

    BAD_OBJECT_REFERENCE = -400
    FEATURE_NOT_SUPPORTED = -120
    INVALID_SESSION = -102
    AUTHENTICATION_FAILED = -100
    CONNECTION_ERROR = 4101
    ACCOUNT_LOCKED = -123


PROJECT_URL: Final[str] = "https://github.com/craigjmidwinter/total-connect-client"

STATUS_URL: Final[str] = "https://status.resideo.com/"

AUTH_CONFIG_ENDPOINT: Final[str] = "https://totalconnect2.com/application.config.json"
AUTH_TOKEN_ENDPOINT: Final[str] = "https://rs.alarmnet.com/TC2API.Auth/token"
HTTP_API_ENDPOINT_BASE: Final[str] = "https://rs.alarmnet.com/TC2API.TCResource/"


def make_http_endpoint(path: str) -> str:
    """Create a full HTTP endpoint URL by joining the base URL with the given path."""
    return urllib.parse.urljoin(HTTP_API_ENDPOINT_BASE, path)


HTTP_API_SESSION_DETAILS_ENDPOINT: Final[str] = make_http_endpoint(
    "api/v3/authentication/sessiondetails"
)
HTTP_API_LOGOUT: Final[str] = make_http_endpoint("api/v3/authentication/logout")



"""Total Connect User."""

import logging
from typing import Any

LOGGER = logging.getLogger(__name__)


class TotalConnectUser:
    """User for Total Connect."""

    def __init__(self, user_info: dict[str, Any]) -> None:
        """Initialize based on UserInfo from LoginAndGetSessionDetails."""
        self._user_id = user_info["UserID"]
        self._username = user_info["Username"]
        self._features = dict(x.split("=") for x in user_info["UserFeatureList"].split(","))
        self._master_user = self._features["Master"] == "1"
        self._user_admin = self._features["User Administration"] == "1"
        self._config_admin = self._features["Configuration Administration"] == "1"

        if self.security_problem():
            LOGGER.warning(
                f"Total Connect user {self._username} has one or "
                "more permissions that are not necessary. Remove "
                "permissions from this user or create a new user "
                "with minimal permissions."
            )

    def security_problem(self) -> bool:
        """Run security checks. Return true if problem."""
        problem = False

        if self._master_user:
            LOGGER.warning(f"User {self._username} is a master user.")
            problem = True

        if self._user_admin:
            LOGGER.warning(f"User {self._username} is a user administrator.")
            problem = True

        if self._config_admin:
            LOGGER.warning(f"User {self._username} is a configuration administrator.")
            problem = True

        return problem

    def __str__(self) -> str:  # pragma: no cover
        """Return a string that is printable."""
        data = (
            f"Username: {self._username}\n"
            f"UserID: {self._user_id}\n"
            f"Master User: {self._master_user}\n"
            f"User Administrator: {self._user_admin}\n"
            f"Configuration Administrator: {self._config_admin}\n"
            "User features:\n"
        )

        for key, value in self._features.items():
            data = data + f"  {key}: {value}\n"

        return data + "\n"
